/*
  $Id: VISD.xs 53 2020-04-09 23:03:57Z abalama $

  Authors:
    Serz Minus (Sergey Lepenkov) L<http://www.serzik.com> E<lt>abalama@cpan.orgE<gt>

  Copyright (C) 1998-2020 D&D Corporation. All Rights Reserved

  This program is free software; you can redistribute it and/or
  modify it under the same terms as Perl itself.

  See LICENSE file
*/

#include <stdlib.h>
#ifndef WIN32
#include <sys/sysinfo.h>
#include <sys/statvfs.h>
#endif

#include "EXTERN.h"
#include "perl.h"
#include "XSUB.h"
#include "VISD.h"

MODULE = VISD        PACKAGE = VISD::SysInfo

PROTOTYPES: ENABLE

#ifdef WIN32
HV*
_getsysinfo()

    PREINIT:
        HV* info_hash;
        AV* info_load;

    CODE:
        info_hash = (HV*)sv_2mortal((SV*)newHV());
        info_load = newAV();

        hv_store(info_hash, "uptime", 6, newSViv(0), 0);
        hv_store(info_hash, "totalram", 8, newSViv(0), 0); // Total usable main memory size
        hv_store(info_hash, "freeram", 7, newSViv(0), 0); // Available memory size
        hv_store(info_hash, "sharedram", 9, newSViv(0), 0); // Amount of shared memory
        hv_store(info_hash, "bufferram", 9, newSViv(0), 0); // Memory used by buffers
        hv_store(info_hash, "totalswap", 9, newSViv(0), 0); // Total swap space size
        hv_store(info_hash, "freeswap", 8, newSViv(0), 0); // Swap space still available
        hv_store(info_hash, "procs", 5, newSViv(0), 0); // Number of current processes
        hv_store(info_hash, "load1", 5, newSVnv(0), 0); // 1 minute load average
        hv_store(info_hash, "load5", 5, newSVnv(0), 0); // 5 minute load average
        hv_store(info_hash, "load15", 6, newSVnv(0), 0); // 15 minute load average

        // Set load
        av_push(info_load, newSVnv(0));
        av_push(info_load, newSVnv(0));
        av_push(info_load, newSVnv(0));
        hv_store(info_hash, "load", 4, newRV_noinc((SV*)info_load), 0); // 1, 5, and 15 minute load averages

        RETVAL = info_hash;

    OUTPUT:
        RETVAL

#else
HV*
_getsysinfo()

    PREINIT:
        struct sysinfo info;
        int ret;
        HV* info_hash;
        AV* info_load;
        double load[3];
        float f_load;

    CODE:
        f_load = 1.f / (1 << SI_LOAD_SHIFT);

        // Get load average
        // See: https://metacpan.org/source/BARABAS/Sys-Load-0.2/Load.xs
        ret = getloadavg(load, 3);
        if(ret == -1) {
            load[0] = 0;
            load[1] = 0;
            load[2] = 0;
        }

        // Get info
        ret = sysinfo(&info);
        info_hash = (HV*)sv_2mortal((SV*)newHV());
        if(ret == -1)
            RETVAL = info_hash;
        else {
            //info_hash = newHV();
            info_load = newAV();

            hv_store(info_hash, "uptime", 6, newSViv(info.uptime), 0);
            hv_store(info_hash, "totalram", 8, newSViv(info.totalram), 0); // Total usable main memory size
            hv_store(info_hash, "freeram", 7, newSViv(info.freeram), 0); // Available memory size
            hv_store(info_hash, "sharedram", 9, newSViv(info.sharedram), 0); // Amount of shared memory
            hv_store(info_hash, "bufferram", 9, newSViv(info.bufferram), 0); // Memory used by buffers
            hv_store(info_hash, "totalswap", 9, newSViv(info.totalswap), 0); // Total swap space size
            hv_store(info_hash, "freeswap", 8, newSViv(info.freeswap), 0); // Swap space still available
            hv_store(info_hash, "procs", 5, newSViv(info.procs), 0); // Number of current processes
            hv_store(info_hash, "load1", 5, newSVnv(info.loads[0] * f_load), 0); // 1 minute load average
            hv_store(info_hash, "load5", 5, newSVnv(info.loads[1] * f_load), 0); // 5 minute load average
            hv_store(info_hash, "load15", 6, newSVnv(info.loads[2] * f_load), 0); // 15 minute load average

            // Set load
            av_push(info_load, newSVnv(load[0]));
            av_push(info_load, newSVnv(load[1]));
            av_push(info_load, newSVnv(load[2]));
            hv_store(info_hash, "load", 4, newRV_noinc((SV*)info_load), 0); // 1, 5, and 15 minute load averages

            //RETVAL = sv_2mortal((SV*)newRV_noinc((SV *)info_hash));
            RETVAL = info_hash;
        }

    OUTPUT:
        RETVAL

#endif

#ifdef WIN32
HV*
_getstatvfs(dir)
    char *dir

    PREINIT:
        HV* st_hash;

    CODE:
        st_hash = (HV*)sv_2mortal((SV*)newHV());

        // Push values as doubles because we don't know size
        hv_store(st_hash, "bsize", 5, newSViv(0), 0);
        hv_store(st_hash, "frsize", 6, newSViv(0), 0);
        hv_store(st_hash, "blocks", 6, newSViv(0), 0);
        hv_store(st_hash, "bfree", 5, newSViv(0), 0);
        hv_store(st_hash, "bavail", 6, newSViv(0), 0);

        // Return
        RETVAL = st_hash;

    OUTPUT:
        RETVAL

#else
HV*
_getstatvfs(dir)
    char *dir

    PREINIT:
        struct statvfs st;
        int ret;
        HV* st_hash;

    CODE:
        // Get data
        ret = statvfs(dir, &st);
        st_hash = (HV*)sv_2mortal((SV*)newHV());
        if (ret == 0) {
            // Push values as doubles because we don't know size
            hv_store(st_hash, "bsize", 5, newSVnv((double)st.f_bsize), 0); // File system block size
            hv_store(st_hash, "frsize", 6, newSVnv((double)st.f_frsize), 0); // Fundamental file system block size
            hv_store(st_hash, "blocks", 6, newSVnv((double)st.f_blocks), 0); // Total number of blocks on file system in units of f_frsize
            hv_store(st_hash, "bfree", 5, newSVnv((double)st.f_bfree), 0); // Total number of free blocks
            hv_store(st_hash, "bavail", 6, newSVnv((double)st.f_bavail), 0); // Number of free blocks available to non-privileged process
        }
        // Return
        RETVAL = st_hash;

    OUTPUT:
        RETVAL

#endif