package VISD::Cetus; # $Id: Cetus.pm 80 2020-04-29 15:34:05Z abalama $
use strict;
use utf8;

=encoding utf8

=head1 NAME

VISD::Cetus - VISD Cetus class

=head1 VERSION

Version 1.00

=head1 SYNOPSIS

    use VISD::Cetus;

=head1 DESCRIPTION

This module provides cetus functionality

=head2 again

Internal use only!

=head2 serializer

Returns serializer object

=head2 check

This handler checks Cetus configuration

=head2 version

This handler prints current version of Cetus

=head1 SEE ALSO

L<VISD>

=head1 AUTHOR

Serż Minus (Sergey Lepenkov) L<http://www.serzik.com> E<lt>abalama@cpan.orgE<gt>

=head1 COPYRIGHT

Copyright (C) 1998-2020 D&D Corporation. All Rights Reserved

=head1 LICENSE

This program is free software; you can redistribute it and/or
modify it under the same terms as Perl itself.

See C<LICENSE> file and L<https://dev.perl.org/licenses/>

=cut

use vars qw/$VERSION/;
$VERSION = '1.00';

use base qw/CTK::App/;

use File::Spec;
use Cwd qw/cwd/;
use List::Util qw/uniq/;
use Sys::Hostname ();
use Try::Tiny;

use CTK::Util qw/ bload bsave /;
use CTK::Serializer;
use CTK::ConfGenUtil;
use CTK::TFVals qw/ :ALL /;

use VISD;
use VISD::Tools;

use constant {
        AREA_WIDTH => 60,
    };

__PACKAGE__->register_handler(
    handler     => "version",
    description => sprintf("%s Version", VISD::PROJECTNAME),
    code => sub {
### CODE:
    my ($self, $meta, @arguments) = @_;
    printf("%s\n", VISD_BANNER);
    return 1;
});

__PACKAGE__->register_handler(
    handler     => "check",
    description => sprintf("%s checking", VISD::PROJECTNAME),
    code => sub {
### CODE:
    my ($self, $meta, @arguments) = @_;
    my $summary = 1;
    printf("Checking %s\n%s\n", VISD_BANNER, "=" x (AREA_WIDTH + 3));

    # CTK status
    {
        _say_name("General status");
        if ($self->status) {
            _say_ok();
        } else {
            $summary = _say_error($self->error);
        }
        $self->debug(sprintf("  Prefix    : %s", $self->prefix));
        $self->debug(sprintf("  Suffix    : %s", $self->suffix));
        $self->debug(sprintf("  Data dir  : %s", $self->datadir));
        $self->debug(sprintf("  Log dir   : %s", $self->logdir));
        $self->debug(sprintf("  Temp dir  : %s", $self->tempdir));
        $self->debug(sprintf("  Root dir  : %s", $self->root));
        $self->debug(sprintf("  Exe dir   : %s", $self->exedir));
        $self->debug(sprintf("  Log file  : %s", $self->logfile));
        $self->debug(sprintf("  Temp file : %s", $self->tempfile));
    }

    # Config status
    {
        _say_name("Config status");
        if ($self->conf("loadstatus")) {
            _say_ok();
        } else {
            $summary = _say_error();
            $self->debug("use --config=/path/to/visd.conf");
        }
        $self->debug(sprintf("  Config file: %s", $self->configfile));
        print _explain($self->config) if $self->verbosemode;
    }

    # RRDs status
    {
        _say_name("RRDs status");
        if (HAVE_RRDTOOLS) {
            _say_ok();
        } else {
            $summary = _say_error();
        }

        _say_name("RRDtool rrdcached status (optional)");
        if (RRDTOOLS_SOCKET) {
            _say_ok();
            $self->debug(sprintf("  RRDtool rrdcached socket: %s", RRDTOOLS_SOCKET));
        } else {
            _say_error();
        }
    }

    # Redis
    my $redis_status = 0;
    {
        my $redis_server = value($self->config('redisserver')) || VISD::REDIS_SERVER;
        _say_name(sprintf("Redis on %s", $redis_server));
        try { # Load Redis module
            require Redis;
            $redis_status = Redis->VERSION;
        };
        unless ($redis_status) {
            $summary = _say_error("Redis module is not installed");
            last;
        }
        my $redis_cnct_to = fv2zero(value($self->config(lc('RedisConnectTimeout')))) || VISD::REDIS_CONNECT_TO;
        my $redis_rqst_to = fv2zero(value($self->config(lc('RedisRequestTimeout')))) || VISD::REDIS_REQUEST_TO;
        my $redis = Redis->new(
            server          => $redis_server,
            name            => "visd_cetus_agent",
            conservative_reconnect => 1,
            cnx_timeout     => $redis_cnct_to,
            read_timeout    => $redis_rqst_to,
            write_timeout   => $redis_rqst_to,
            no_auto_connect_on_new => 1,
        );
        try {
            $redis->connect();
        } catch {
            $redis_status = 0;
            $summary = _say_error($_);
        };
        last unless $redis_status;
        _say_ok();
        print _explain($redis->info) if $self->verbosemode;
    }

    # Get AHosts
    {
        VISD::Tools::set_fetch_timeout($self->config("timeout"));
        my $conf_hosts = $self->config("host");
        if (is_hash($conf_hosts)) {
            foreach my $host (keys %$conf_hosts) {
                my $source_type = lc(uv2null(value($conf_hosts, $host, "sourcetype")));
                if ($source_type ne "remote") { # Redis
                    _say_name("Host $host (redis)");
                    $redis_status ? _say_ok() : _say_error("Redis unavailable");
                    next;
                }
                # Remote
                my $url = uv2null(value($conf_hosts, $host, "sourceurl"));
                _say_name("Host $host (remote)");
                my ($status, $msg) = fetch($url); # Fetching data;
                if ($status) { # Data is loaded!
                    my $data = $self->serializer->deserialize($msg);
                    if ($self->serializer->status) {
                        _say_ok();
                        $self->debug(sprintf("  URL: %s", $url));
                        print _explain($data) if $self->verbosemode;
                    } else {
                        $summary = _say_error($self->serializer->error);
                    }
                } else {
                    $summary = _say_error($msg);
                }
            }
        }
    }

    # Summary status
    printf "%s\nSummary status:", "=" x (AREA_WIDTH + 3);
    $summary ? _say_ok() : _say_error();
    return $summary;
});

__PACKAGE__->register_handler(
    handler     => "graph",
    description => "Graphs plotting",
    code => sub {
### CODE:
    my ($self, $meta, @arguments) = @_;
    my $graphsdir = shift(@arguments) || $self->conf("graphsdir") || cwd();
    unless (-e $graphsdir or -d _ or -l $graphsdir) {
        $self->error(sprintf("Can't use \"%s\" directory for graphs plotting. The directory not exists", $graphsdir));
        return 0;
    }
    unless (-w $graphsdir) {
        $self->error(sprintf("Can't use \"%s\" directory for graphs plotting. Permission denied", $graphsdir));
        return 0;
    }

    # Get last data from file
    my $ldf = File::Spec->catfile($self->datadir(), VISD::LAST_DATA_FILE);
    unless (-e $ldf) {
        $self->error(sprintf("File \"%s\" not found", $ldf));
        return 0;
    }
    my $data = {};
    if (my $bindata = bload($ldf, 1)) {
        $data = $self->serializer->deserialize($bindata);
        unless ($self->serializer->status) {
            $self->error($self->serializer->error);
            return 0;
        }
    } else {
        $self->error(sprintf("Can't loading file \"%s\"", $ldf));
        return 0;
    }

    # Get ahosts & agraphs from last data
    my $hostname = value($data, "hostname") // DEFAULT_HOSTNAME;
    my $ahosts = array($data, "ahosts");
    push @$ahosts, $hostname;
    my $agraphs = hash($data, "agraphs");
    my $net_interfaces = hash($data, "net/interfaces");
    my $fs_dev = hash($data, "fs/dev");
    $agraphs->{$hostname} = {
        interfaces => [keys(%$net_interfaces)],
        devices => [keys(%$fs_dev)],
    };

    # Get hosts for processing
    my @hosts = req_or_avail(array($self->option("host")), $ahosts);

    # Get groups
    my @groups = req_or_avail(array($self->option("group")), VISD::Tools::AVAILABLE_GROUPS);

    # Get zooms
    my @zooms = req_or_avail(array($self->option("zoom")), VISD::Tools::AVAILABLE_ZOOMS);

    # Get Req Names
    my @req_names = @{ array($self->option("name")) };

    # Debug
    #if ($self->debugmode) {
    #    printf "GraphsDir   : %s\n", $graphsdir;
    #    printf "Hosts       : %s\n", join(", ", @hosts ? @hosts : (qw/none/) );
    #    printf "Groups      : %s\n", join(", ", @groups ? @groups : (qw/none/) );
    #    printf "Zooms       : %s\n", join(", ", @zooms ? @zooms : (qw/none/) );
    #}

    foreach my $host (@hosts) {
        foreach my $group (@groups) {
            my @names = (""); # Default. No need name;
            if ($group eq 'net') {
                @names = req_or_avail(array($self->option("name")), array($agraphs, $host, "interfaces"));
            } elsif ($group =~ /^dev/) {
                @names = req_or_avail(array($self->option("name")), array($agraphs, $host, "devices"));
            }
            foreach my $name (@names) {
                foreach my $zoom (@zooms) {
                    my $rrdf = File::Spec->catfile($self->datadir(), lc($host), get_rrdfn($group, $name));
                    my $pngf = File::Spec->catfile($graphsdir, get_pngfn($host, $group, $name, $zoom));
                    printf "Plotting %s... ", $pngf if $self->debugmode;
                    my ($rrdstatus, $rrdmsg) = rrd_graph(
                        file => $rrdf,
                        group => $group,
                        zoom => $zoom,
                    );
                    if ($rrdstatus) {
                        #$self->logger->log_info("RRD database for %s host was successfully created (%s)", $host, $rrd_sys_file);
                        if (bsave($pngf, $rrdmsg)) {
                            print "ok\n" if $self->debugmode;
                        } else {
                            print "fail\n" if $self->debugmode;
                            $self->log_error("Can't save data to file \"%s\"", $pngf);
                            printf STDERR "Can't save data to file \"%s\"\n", $pngf if $self->debugmode;
                        }
                    } else {
                        print "fail\n" if $self->debugmode;
                        $self->log_error("Can't plot RRD graph %s: %s", $pngf, $rrdmsg);
                        printf STDERR "Can't plot RRD graph %s: %s\n", $pngf, $rrdmsg if $self->debugmode;
                    }
                }
            }
        }
    }
    return 1;
});

sub again {
    my $self = shift;
    $self->{serializer} = new CTK::Serializer(VISD::SERIALIZER_FORMAT, attrs => VISD::SERIALIZER_ATTRS);
    return $self->SUPER::again();
}

sub serializer { shift->{serializer} }

sub _explain {
    require Data::Dumper;
    my $dumper = Data::Dumper->new ( [shift] );
    $dumper->Indent(1)->Terse(1);
    $dumper->Sortkeys(1) if $dumper->can("Sortkeys");
    return $dumper->Dump;
}
sub _say_name {
    my $msg = shift // '';
    my $al = AREA_WIDTH - length($msg);
    if ($al > 0) {
        print $msg, "." x $al;
    } else {
        print $msg;
    }
}
sub _say_ok {
    print " ok\n";
    return 1;
}
sub _say_error {
    my $msg = shift // '';
    print " error\n";
    printf STDERR "%s\n", $msg if length $msg;
    return 0;
}

1;

__END__
