package VISD::Cetus::Agent; # $Id: Agent.pm 80 2020-04-29 15:34:05Z abalama $
use strict;
use utf8;

=encoding utf8

=head1 NAME

VISD::Cetus::Agent - VISD Cetus Daemon

=head1 VERSION

Version 1.00

=head1 SYNOPSIS

    use VISD::Cetus::Agent;

=head1 DESCRIPTION

This module provides running the cetus daemon

=head2 foo

This handler for developmenters only

=head2 get

This handler runs getting of data from sources periodical (1 call per one second)

=head1 METHODS

List of methods

=head2 new

    my $daemon = new VISD::Cetus::Agent('foo', (
        debug       => debugmode,
        loglevel    => $loglevel,
        forks       => 3,
    ));

=head2 devices

    my @devices = $daemon->devices( qw/ foo bar baz / );

Returns allowed devices list

=head2 error

    my $error = $daemon->error("new error string");
    my $error = $daemon->error;

Gets or sets the last error string. Returns last error string

=head2 genhkeyf

    my $key = $daemon->genhkeyf("hostname", "groupname", "keyname");

Returns full key string for specified hostname

=head2 genkeyf

    my $local_key = $daemon->genkeyf("groupname", "keyname");
    my $local_key = $daemon->genkeyf("keyname");

Returns full key string for current hostname


=head2 get_list_handlers

    my @list = get_list_handlers();

Returns list of hashes that contain metadata of each handlers without "handler" value

=head2 get_remote_data

    if (my $data = $daemon->get_remote_data($hostname, $expire)) {
        # ... data contents structure ...
    }

Returns structure that was loaded from remote sources

=head2 get_status

    my $status = get_status("key", {...data..});

Returns calculated status by data's values

=head2 init

This is internal method of the CTK::Daemon context

=head2 interfaces

    my @interfaces = $daemon->interfaces( qw/ foo bar baz / );

Returns allowed interfaces list

=head2 lookup_handler

    my $handler = lookup_handler("FOO");

Look's the handler by name; returns handler or undef

=head2 redis

    my $redis = $daemon->redis();

Returns redis object or undef

=head2 register_handler

This function (not method) register new handler for job processing

=head2 run

This is internal method of the CTK::Daemon context

=head2 serializer

    my $sr = $daemon->serializer();

Returns serializer object

=head1 SEE ALSO

L<VISD>

=head1 AUTHOR

Serż Minus (Sergey Lepenkov) L<http://www.serzik.com> E<lt>abalama@cpan.orgE<gt>

=head1 COPYRIGHT

Copyright (C) 1998-2020 D&D Corporation. All Rights Reserved

=head1 LICENSE

This program is free software; you can redistribute it and/or
modify it under the same terms as Perl itself.

See C<LICENSE> file and L<https://dev.perl.org/licenses/>

=cut

use vars qw/$VERSION $REDIS_VERSION/;
$VERSION = '1.00';

use base qw/CTK::Daemon/;

use Module::Loaded qw/is_loaded/;
use Try::Tiny;

use AnyEvent;
use Time::HiRes qw/ gettimeofday /;

use Carp;
use CTK::ConfGenUtil;
use CTK::TFVals qw/ :ALL /;
use CTK::Util qw/ touch bsave preparedir isTrueFlag /;
use CTK::Serializer;

# General (default)
use Sys::Hostname ();
use File::Spec ();
use Text::ParseWords qw/ shellwords /;

use VISD;
use VISD::SysInfo;
use VISD::Tools;

use constant {
    FORKS               => 1,
    START_DELAY         => 3,
    INTERVAL            => 1,
    COMMON_KEYS         => [qw/
        status
        ctime uptime processes
        cpuusage idle iodelay iowait
    /],

    # Expire time
    EXPIRE_GENERAL      => 60*5, # 5 min
    EXPIRE_ROUTING      => 60*60, # 1 hour
    EXPIRE_FS           => 60*20, # 20 min

    # Task intervals
    INTERVAL_GENERAL    => 60, # 1 min
    INTERVAL_SYS        => 20, # 20 sec
    INTERVAL_NET        => 30, # 30 sec
    INTERVAL_NET_ROUTE  => 60*20, # 20 min
    INTERVAL_DEV        => 60*5, # 5 min

    # System health statuses
    STATUS_UNKNOWN      => -2, # Unknown
    STATUS_WARNING      => -1, # Warning
    STATUS_CRITICAL     => 0, # Critical
    STATUS_GOOD         => 1, # Good

    # limits
    DEVICE_WRL          => 90, # 90%
    DEVICE_CRL          => 99, # 99%
    CPUUSAGE_WRL        => 90, # 90%
    CPUUSAGE_CRL        => 99, # 99%
    LOADAVG5_WRL        => 1.0,
    LOADAVG5_CRL        => 1.2,
};

# Load Redis module
$REDIS_VERSION = 0;
try {
    require Redis;
    $REDIS_VERSION = Redis->VERSION;
};

my $order = 0;
my %handlers = (); # Handlers

__PACKAGE__->register_handler( # perl -Iblib/arch -Ilib bin/cetus -c conf/visd.conf -d start
    name        => "default",
    description => "General processing",
    order       => $order++,
    interval    => INTERVAL_GENERAL, # 1 min
    handler     => sub {
    my ($self, $iter) = @_;
    my $sr = $self->serializer();
    my @ahosts = @{( $self->{ahosts} )};
    my $ldf = File::Spec->catfile($self->ctk->datadir(), VISD::LAST_DATA_FILE);
    my $sattr = VISD::SERIALIZER_ATTRS->{VISD::SERIALIZER_FORMAT}->[0];
    my %pretty = %$sattr; $pretty{pretty} = 1;
    my %common_data = (
        iter            => $iter,
        version         => VISD_FULL_VERSION,
        banner          => VISD_BANNER,
    );
    my $common_keys = COMMON_KEYS;

    #$self->interrupt(1) if $iter > 1;

    # Skip first call
    return 1 if $iter <= 1;

    ##
    ## Get last data from Redis
    ##
    my %last_data = (
        %common_data,
        hostname        => $self->{hostname},
        ahosts          => [@ahosts],
    );
    if (my $redis = $self->redis) {
        # General status
        $redis->set($self->genkeyf("status"), $self->get_status("general", {
            status_sys => $redis->get($self->genkeyf(status => "sys")),
            status_net => $redis->get($self->genkeyf(status => "net")),
            status_dev => $redis->get($self->genkeyf(status => "dev")),
        }));

        # General data
        foreach my $k (@$common_keys) {
            my $v = fv2zero($redis->get($self->genkeyf($k)));
            $last_data{$k} = $v * 1.0;
        }

        # Kernel Version & Machine
        $last_data{kversion} = $redis->get($self->genkeyf("kversion")) || $self->{kversion};
        $last_data{kmachine} = $redis->get($self->genkeyf("kmachine")) || $self->{kmachine};

        # CPU info
        my $cpuinfo_j = $redis->get($self->genkeyf("cpuinfo"));
        $last_data{cpuinfo} = $sr->deserialize( $cpuinfo_j ) if $cpuinfo_j;

        # LoadAVG
        my $la_j = $redis->get($self->genkeyf("loadavg"));
        $last_data{loadavg} = $sr->deserialize( $la_j ) if $la_j;

        # Memory
        my $mem_j = $redis->get($self->genkeyf("memory"));
        $last_data{memory} = $sr->deserialize( $mem_j ) if $mem_j;

        # Swap
        my $swap_j = $redis->get($self->genkeyf("swap"));
        $last_data{swap} = $sr->deserialize( $swap_j ) if $swap_j;

        # Net
        my $iface_j = $redis->get($self->genkeyf(net => "interfaces"));
        my $rout4_j = $redis->get($self->genkeyf(net => "route.ipv4"));
        my $rout6_j = $redis->get($self->genkeyf(net => "route.ipv6"));
        $last_data{net} = {
            interfaces => $iface_j ? $sr->deserialize( $iface_j ) : {},
            route => {
                ipv4 => $rout4_j ? $sr->deserialize( $rout4_j ) : [],
                ipv6 => $rout6_j ? $sr->deserialize( $rout6_j ) : [],
            },
        };

        # Filesystems
        my $fs_j = $redis->get($self->genkeyf(fs => "dev"));
        $last_data{fs} = {
            dev => $fs_j ? $sr->deserialize( $fs_j ) : {},
        };
    }

    ##
    ## Get last data from addition hosts
    ##
    my %agrpaphs;
    foreach my $ahost (@ahosts) {
        my $source_type = lc(uv2null(value($self->ctk->config("host"), $ahost, "sourcetype")));
        my ($astatus, $ifaces, $devs) = (STATUS_UNKNOWN, {}, {});
        if ($source_type eq "remote") {
            if (my $remote_data = $self->get_remote_data($ahost, INTERVAL_GENERAL)) { # Data is loaded!
                $astatus = value($remote_data, "status");
                $ifaces = hash($remote_data, "net/interfaces");
                $devs = hash($remote_data, "fs/dev");
            }
        } else { # redis (default)
            if (my $redis = $self->redis) {
                my $hldf = File::Spec->catfile($self->ctk->datadir(), lc($ahost), VISD::LAST_DATA_FILE);
                my %hl_data = (
                    %common_data,
                    hostname        => $ahost,
                    ahosts          => [],
                );

                # General data
                foreach my $k (@$common_keys) {
                    $hl_data{$k} = fv2zero($redis->get($self->genhkeyf($ahost, general => $k))) * 1.0;
                }
                $astatus = $hl_data{status};

                # Kernel Version & Machine
                $hl_data{kversion} = uv2null($redis->get($self->genhkeyf($ahost, general => "kversion")));
                $hl_data{kmachine} = uv2null($redis->get($self->genhkeyf($ahost, general => "kmachine")));

                # CPU info
                my $cpuinfo_j = $redis->get($self->genhkeyf($ahost, general => "cpuinfo"));
                $hl_data{cpuinfo} = $sr->deserialize( $cpuinfo_j ) if $cpuinfo_j;

                # LoadAVG
                my $la_j = $redis->get($self->genhkeyf($ahost, general => "loadavg"));
                $hl_data{loadavg} = $sr->deserialize( $la_j ) if $la_j;

                # Memory
                my $mem_j = $redis->get($self->genhkeyf($ahost, general => "memory"));
                $hl_data{memory} = $sr->deserialize( $mem_j ) if $mem_j;

                # Swap
                my $swap_j = $redis->get($self->genhkeyf($ahost, general => "swap"));
                $hl_data{swap} = $sr->deserialize( $swap_j ) if $swap_j;

                # Net
                my $iface_j = $redis->get($self->genhkeyf($ahost, net => "interfaces"));
                   $ifaces = $iface_j ? $sr->deserialize( $iface_j ) : {};
                my $rout4_j = $redis->get($self->genhkeyf($ahost, net => "route.ipv4"));
                my $rout6_j = $redis->get($self->genhkeyf($ahost, net => "route.ipv6"));
                $hl_data{net} = {
                    interfaces => $ifaces,
                    route => {
                        ipv4 => $rout4_j ? $sr->deserialize( $rout4_j ) : [],
                        ipv6 => $rout6_j ? $sr->deserialize( $rout6_j ) : [],
                    },
                };

                # Filesystems
                my $fs_j = $redis->get($self->genhkeyf($ahost, fs => "dev"));
                   $devs = $fs_j ? $sr->deserialize( $fs_j ) : {};
                $hl_data{fs} = {
                    dev => $devs,
                };

                # Try save all data to file
                unless (bsave($hldf, $sr->serialize( VISD::SERIALIZER_FORMAT, \%hl_data, \%pretty), 1)) {
                    $self->error(sprintf("Can't save host data to file \"%s\"", $hldf));
                }
            }
        }
        $agrpaphs{$ahost} = {
            status => $astatus,
            interfaces => [keys(%$ifaces)],
            devices => [keys(%$devs)],
        };
    }
    $last_data{agraphs} = \%agrpaphs;

    ##
    ## Save all last data to file
    ##
    unless (bsave($ldf, $sr->serialize( VISD::SERIALIZER_FORMAT, \%last_data, \%pretty), 1)) {
        $self->error(sprintf("Can't save data to file \"%s\"", $ldf));
        $self->exception(1);
        return 0;
    }

    return 1;
});

__PACKAGE__->register_handler( # perl -Iblib/arch -Ilib bin/cetus -c conf/visd.conf -d get -k start
    name        => "get",
    description => "Getting data from general sources",
    order       => $order++,
    interval    => INTERVAL_SYS, # 20 sec
    handler     => sub {
    my ($self, $iter) = @_;
    #my $client = $self->client;

    ##
    ## Get general system data
    ##
    my %stat = get_stat();
    my %sysinfo = get_sysinfo();
    if (my $redis = $self->redis) { # Store to redis
        #return $self->interrupt(1); # Redis required for Cetus working
        #$redis->setex("test.bar", 5, scalar(localtime(time())));

        $redis->setex($self->genkeyf("ctime"), EXPIRE_GENERAL, $stat{ctime} || 0);
        $redis->setex($self->genkeyf("uptime"), EXPIRE_GENERAL, $sysinfo{uptime} || 0);
        $redis->setex($self->genkeyf("processes"), EXPIRE_GENERAL, $sysinfo{procs} || 0);

        # CPU
        $redis->setex($self->genkeyf("cpuusage"), EXPIRE_GENERAL, $stat{cpuusage} || 0);
        $redis->setex($self->genkeyf("idle"), EXPIRE_GENERAL, $stat{idle} || 0);
        $redis->setex($self->genkeyf("iodelay"), EXPIRE_GENERAL, $stat{iodelay} || 0);
        $redis->setex($self->genkeyf("iowait"), EXPIRE_GENERAL, $stat{iowait} || 0);

        # LoadAVG
        my $avgs = $sysinfo{load} || [0,0,0];
        $redis->setex($self->genkeyf("loadavg"), EXPIRE_GENERAL,
            $self->serializer->serialize( $avgs ),
        );

        # Swap
        $redis->setex($self->genkeyf("swap"), EXPIRE_GENERAL, $self->serializer->serialize({
            free    => $sysinfo{freeswap} || 0,
            used    => $sysinfo{usedswap} || 0,
            total   => $sysinfo{totalswap} || 0,
        }));

        # Memory
        $redis->setex($self->genkeyf("memory"), EXPIRE_GENERAL, $self->serializer->serialize({
            free    => $sysinfo{freeram} || 0,
            used    => $sysinfo{usedram} || 0,
            total   => $sysinfo{totalram} || 0,
            shared  => $sysinfo{sharedram} || 0,
            buffer  => $sysinfo{bufferram} || 0,
        }));

        # Status
        my $procs = ($self->{cpuinfo}->{cpus} || 0) * ($self->{cpuinfo}->{sockets} || 0);
        $redis->setex($self->genkeyf(status => "sys"), EXPIRE_GENERAL, $self->get_status("sys", {
            procs   => $procs || 1,
            avg1    => ($avgs->[0] || 0) * 1.0,
            avg5    => ($avgs->[1] || 0) * 1.0,
            avg15   => ($avgs->[2] || 0) * 1.0,
            cpu     => ($stat{cpuusage} || 0) * 100,
            mem     => $sysinfo{totalram} ? (100 * ($sysinfo{usedram} || 0) / $sysinfo{totalram}) : 0,
            swp     => $sysinfo{totalswap} ? (100 * ($sysinfo{usedswap} || 0) / $sysinfo{totalswap}) : 0,
            iod     => ($stat{iodelay} || 0) * 100,
        }));

        #$self->logger->log_debug("Default say [%d]:\n%s", $iter, VISD::explain(\%general));
        #$self->logger->log_debug(VISD::explain($self));
    }

    ##
    ## RRD initialize & updating
    ##
    if ($self->{use_graphs} && HAVE_RRDTOOLS) {
        my $ahosts = $self->{ahosts};
        foreach my $host ($self->{hostname}, @$ahosts) { # AHosts
            my $dir = File::Spec->catdir($self->ctk->datadir(), lc($host));
            my $rrd_sys_file = File::Spec->catfile($dir, "sys.rrd");

            # Create RRD database
            unless (-e $rrd_sys_file) {
                my ($rrdstatus, $rrdmsg) = rrd_create(
                    file => $rrd_sys_file,
                    type => "sys",
                    step => INTERVAL_SYS,
                );
                if ($rrdstatus) {
                    $self->logger->log_info("RRD database for %s host was successfully created (%s)", $host, $rrd_sys_file);
                } else {
                    $self->logger->log_error("Can't create RRD database for %s host (%s): %s", $host, $rrd_sys_file, $rrdmsg);
                }
                next;
            }

            # Update RRD database
            my %data = ();
            if ($host eq $self->{hostname}) {
                # Get local data
                my $avgs = $sysinfo{load} || [0,0,0];
                %data = (
                    cpu => $stat{cpuusage} ? sprintf("%.0f", $stat{cpuusage} * 100.0) * 1 : undef,
                    mem => $sysinfo{totalram} ? sprintf("%.0f", 100.0 * ($sysinfo{usedram} || 0.0) / $sysinfo{totalram}) : undef,
                    swp => $sysinfo{totalswap} ? sprintf("%.0f", 100.0 * ($sysinfo{usedswap} || 0.0) / $sysinfo{totalswap}) : undef,
                    iod => $stat{iodelay} ? sprintf("%.0f", $stat{iodelay} * 100.0) * 1 : undef,
                    avg1    => ($avgs->[0] || 0) * 1.0,
                    avg5    => ($avgs->[1] || 0) * 1.0,
                    avg15   => ($avgs->[2] || 0) * 1.0,
                );
            } else {
                # Get remote data
                if (lc(uv2null(value($self->ctk->config("host"), $host, "sourcetype"))) eq "remote") {
                    if (my $remote_data = $self->get_remote_data($host, INTERVAL_SYS)) { # Data is loaded!
                        my $avgs = array($remote_data, "loadavg");
                        %data = (
                            cpu => value($remote_data, "cpuusage") ? sprintf("%.0f", value($remote_data, "cpuusage") * 100.0) * 1 : undef,
                            mem => value($remote_data, "memory/total") ? sprintf("%.0f", 100.0 * (value($remote_data, "memory/used") || 0.0) / value($remote_data, "memory/total")) : undef,
                            swp => value($remote_data, "swap/total") ? sprintf("%.0f", 100.0 * (value($remote_data, "swap/user") || 0.0) / value($remote_data, "swap/total")) : undef,
                            iod => value($remote_data, "iodelay") ? sprintf("%.0f", value($remote_data, "iodelay") * 100.0) * 1 : undef,
                            avg1    => ($avgs->[0] || 0) * 1.0,
                            avg5    => ($avgs->[1] || 0) * 1.0,
                            avg15   => ($avgs->[2] || 0) * 1.0,
                        );
                    }
                } else { # redis (default)
                    if (my $redis = $self->redis) {
                        my $j_memory = $redis->get($self->genhkeyf($host, general => "memory"));
                        my $mem = hash($j_memory ? $self->serializer->deserialize($j_memory) : {});
                        my $j_swap = $redis->get($self->genhkeyf($host, general => "swap"));
                        my $swp = hash($j_swap ? $self->serializer->deserialize($j_swap) : {});
                        my $_cpu = $redis->get($self->genhkeyf($host, general => "cpuusage"));
                        my $_iod = $redis->get($self->genhkeyf($host, general => "iodelay"));
                        my $j_loadavg = $redis->get($self->genhkeyf($host, general => "loadavg"));
                        my $avgs = array($j_loadavg ? $self->serializer->deserialize($j_loadavg) : []);
                        %data = (
                            cpu => $_cpu ? sprintf("%.0f", $_cpu * 100.0) * 1 : undef,
                            iod => $_iod ? sprintf("%.0f", $_iod * 100.0) * 1 : undef,
                            mem => value($mem, "total") ? sprintf("%.0f", 100.0 * (value($mem, "used") || 0.0) / value($mem, "total")) : undef,
                            swp => value($swp, "total") ? sprintf("%.0f", 100.0 * (value($swp, "used") || 0.0) / value($swp, "total")) : undef,
                            avg1    => ($avgs->[0] || 0) * 1.0,
                            avg5    => ($avgs->[1] || 0) * 1.0,
                            avg15   => ($avgs->[2] || 0) * 1.0,
                        );
                    }
                }
            }
            my ($rrdstatus, $rrdmsg) = rrd_update(
                file => $rrd_sys_file,
                type => "sys",
                %data
            );
            unless ($rrdstatus) {
                $self->logger->log_info("Can't update RRD database %s: %s", $rrd_sys_file, $rrdmsg);
                next;
            }
        }
    }

    #$self->logger->log_debug("Foo say [%d]", $iter);
    #$self->logger->log_debug(Dumper([get_list_handlers()]));

    #$self->interrupt(1);
    #$self->hangup(1) if $iter >= 3; # Reloading
    #$self->error("Oops :-) Cetus!");

    return 1;
});

__PACKAGE__->register_handler( # perl -Iblib/arch -Ilib bin/cetus -c conf/visd.conf -d get_net -k start
    name        => "get_net",
    description => "Getting network state data from sources",
    order       => $order++,
    interval    => INTERVAL_NET, # 30 sec
    handler     => sub {
    my ($self, $iter) = @_;

    ##
    ## Get network state data
    ##
    my %in_net_dev = get_net_dev(); # Interfaces
    my %net_dev;
    foreach my $iface ($self->interfaces(keys(%in_net_dev))) {
        $net_dev{$iface} = $in_net_dev{$iface};
    }

    # Store data to Redis
    if (my $redis = $self->redis) {
        # Status
        $redis->setex($self->genkeyf(status => "net"), EXPIRE_GENERAL, $self->get_status("net", \%net_dev));

        # Store data
        $redis->setex($self->genkeyf(net => "interfaces"), EXPIRE_GENERAL,
            $self->serializer->serialize( {%net_dev} ),
        );
    }

    ##
    ## RRD initialize & updating
    ##
    if ($self->{use_graphs} && HAVE_RRDTOOLS) {
        my $ahosts = $self->{ahosts};
        foreach my $host ($self->{hostname}, @$ahosts) { # AHosts
            my $dir = File::Spec->catdir($self->ctk->datadir(), lc($host));
            my $data = {};
            if ($host eq $self->{hostname}) {
                # Get local data
                $data = \%net_dev;
            } else {
                # Get remote date
                if (lc(uv2null(value($self->ctk->config("host"), $host, "sourcetype"))) eq "remote") {
                    if (my $remote_data = $self->get_remote_data($host, INTERVAL_NET)) { # Data is loaded!
                        $data = hash($remote_data, "net/interfaces");
                    }
                } else { # redis (default)
                    if (my $redis = $self->redis) {
                        my $json = $redis->get($self->genhkeyf($host, net => "interfaces")) // "";
                        $data = hash($json ? $self->serializer->deserialize($json) : {});
                        #$self->logger->log_info("Key = %s; Net = %s", $json);
                    }
                }
            }
            my @interfaces = keys %$data;
            foreach my $iface (@interfaces) {
                my $rrd_net_file = File::Spec->catfile($dir, sprintf("net_%s.rrd", $iface));

                # Create RRD database
                unless (-e $rrd_net_file) {
                    my ($rrdstatus, $rrdmsg) = rrd_create(
                        file => $rrd_net_file,
                        type => "net",
                        step => INTERVAL_NET,
                    );
                    if ($rrdstatus) {
                        $self->logger->log_info("RRD database for %s host was successfully created (%s)", $host, $rrd_net_file);
                    } else {
                        $self->logger->log_error("Can't create RRD database for %s host (%s): %s", $host, $rrd_net_file, $rrdmsg);
                    }
                    next;
                }

                # Update RRD database
                my ($rrdstatus, $rrdmsg) = rrd_update(
                    file => $rrd_net_file,
                    type => "net",
                    ## data
                    in => value($data, $iface, "receive") || undef,
                    out => value($data, $iface, "transmit") || undef,
                );
                unless ($rrdstatus) {
                    $self->logger->log_info("Can't update RRD database %s: %s", $rrd_net_file, $rrdmsg);
                    next;
                }
            }
        }
    }


    return 1;
});

__PACKAGE__->register_handler( # perl -Iblib/arch -Ilib bin/cetus -c conf/visd.conf -d get_net_route -k start
    name        => "get_net_route",
    description => "Getting routing table from sources",
    order       => $order++,
    interval    => INTERVAL_NET_ROUTE, # 20 min
    handler     => sub {
    my ($self, $iter) = @_;

    if (my $redis = $self->redis) {
        my @ipv4_route = get_net_route();
        my @ipv6_route = get_net_ipv6_route();
        $redis->setex($self->genkeyf(net => "route.ipv4"), EXPIRE_ROUTING,
            $self->serializer->serialize( [@ipv4_route] ),
        );
        $redis->setex($self->genkeyf(net => "route.ipv6"), EXPIRE_ROUTING,
            $self->serializer->serialize( [@ipv6_route] ),
        );
    }

    return 1;
});

__PACKAGE__->register_handler( # perl -Iblib/arch -Ilib bin/cetus -c conf/visd.conf -d get_dev -k start
    name        => "get_dev",
    description => "Getting filesystem data from sources",
    order       => $order++,
    interval    => INTERVAL_DEV, # 5 min
    handler     => sub {
    my ($self, $iter) = @_;

    ##
    ## Get devices state data
    ##
    my %in_devices = get_disk_free(); # Devices
    my %devices;
    foreach my $dev ($self->devices(keys(%in_devices))) {
        $devices{$dev} = $in_devices{$dev};
    }

    # Store data to Redis
    if (my $redis = $self->redis) {
        # Status
        $redis->setex($self->genkeyf(status => "dev"), EXPIRE_FS, $self->get_status("dev", \%devices));

        # Store data
        $redis->setex($self->genkeyf(fs => "dev"), EXPIRE_FS,
            $self->serializer->serialize( {%devices} ),
        );
    }

    ##
    ## RRD initialize & updating
    ##
    if ($self->{use_graphs} && HAVE_RRDTOOLS) {
        my $ahosts = $self->{ahosts};
        foreach my $host ($self->{hostname}, @$ahosts) { # AHosts
            my $dir = File::Spec->catdir($self->ctk->datadir(), lc($host));
            my $data = {};
            if ($host eq $self->{hostname}) {
                # Get local data
                $data = \%devices;
            } else {
                # Get remote date
                if (lc(uv2null(value($self->ctk->config("host"), $host, "sourcetype"))) eq "remote") {
                    if (my $remote_data = $self->get_remote_data($host, INTERVAL_DEV)) { # Data is loaded!
                        $data = hash($remote_data, "fs/dev");
                    }
                } else { # redis (default)
                    if (my $redis = $self->redis) {
                        my $json = $redis->get($self->genhkeyf($host, fs => "dev")) // "";
                        $data = hash($json ? $self->serializer->deserialize($json) : {});
                    }
                }
            }
            my @devs = keys %$data;
            foreach my $dev (@devs) {
                my $rrd_dev_file = File::Spec->catfile($dir, sprintf("fs_%s.rrd", dev2name($dev)));

                # Create RRD database
                unless (-e $rrd_dev_file) {
                    my ($rrdstatus, $rrdmsg) = rrd_create(
                        file => $rrd_dev_file,
                        type => "dev",
                        step => INTERVAL_DEV, # 5 min
                    );
                    if ($rrdstatus) {
                        $self->logger->log_info("RRD database for %s host was successfully created (%s)", $host, $rrd_dev_file);
                    } else {
                        $self->logger->log_error("Can't create RRD database for %s host (%s): %s", $host, $rrd_dev_file, $rrdmsg);
                    }
                    next;
                }

                # Update RRD database
                my $dd = hash($data->{$dev});
                my $total = value($dd, "total");
                my $used = value($dd, "used");
                my ($rrdstatus, $rrdmsg) = rrd_update(
                    file => $rrd_dev_file,
                    type => "dev",
                    ## data
                    usage => $total ? sprintf("%.0f", 100.0 * ($used || 0.0) / $total) : undef,
                    read => value($dd, "diskread") || undef,
                    write => value($dd, "diskwrite") || undef,
                );
                #$self->logger->log_info("T: %s; U: %s; R: %s; W: %s", $total // '', $used // '', value($dd, "diskwrite") // '', value($dd, "diskwrite") // '');
                unless ($rrdstatus) {
                    $self->logger->log_info("Can't update RRD database %s: %s", $rrd_dev_file, $rrdmsg);
                    next;
                }

            }
        }
    }

    return 1;
});


sub new {
    my $class = shift;
    my $name = shift;
    my %daemon_options = @_;
    $daemon_options{forks} //= FORKS;
    my $self = $class->SUPER::new($name, %daemon_options);
    $self->{error} = '';
    $self->{command} = $daemon_options{command} // "default";
    $self->{hostname} = Sys::Hostname::hostname() // DEFAULT_HOSTNAME;
    $self->{redis} = undef;

    # Set Graphs
    my $graphs = $self->ctk->config("graphs") // "enabled";
    $self->{use_graphs} = isTrueFlag($graphs);

    # Set AHosts
    my $conf_hosts = $self->ctk->config("host");
    my @ahosts = ();
    push @ahosts, keys %$conf_hosts if is_hash($conf_hosts);
    $self->{ahosts} = [@ahosts];

    # Set allowed interfaces
    my $allowedinterfaces = $self->ctk->config("allowedinterfaces");
    $self->{allowedinterfaces} = [ shellwords( @{(array($allowedinterfaces))} ) ] if $allowedinterfaces;

    # Set allowed devices
    my $alloweddevices = $self->ctk->config("alloweddevices");
    $self->{alloweddevices} = [ shellwords( @{(array($alloweddevices))} ) ] if $alloweddevices;

    # Set serializer
    $self->{serializer} = new CTK::Serializer(VISD::SERIALIZER_FORMAT, attrs => VISD::SERIALIZER_ATTRS);

    ##
    ## Set limits
    ##
    # DeviceWRL -- warning reference limit (WRL)
    my $devicewrl = value($self->ctk->config(), "devicewrl") // DEVICE_WRL;
    $self->{devicewrl} = is_int($devicewrl) ? $devicewrl : DEVICE_WRL;

    # DeviceCRL -- critical reference limit (CRL)
    my $devicecrl = value($self->ctk->config(), "devicecrl") // DEVICE_CRL;
    $self->{devicecrl} = is_int($devicecrl) ? $devicecrl : DEVICE_CRL;

    # CPUUsageWRL & CPUUsageCRL
    my $cpuusagewrl = value($self->ctk->config(), "cpuusagewrl") // CPUUSAGE_WRL;
    $self->{cpuusagewrl} = is_int($cpuusagewrl) ? $cpuusagewrl : CPUUSAGE_WRL;
    my $cpuusagecrl = value($self->ctk->config(), "cpuusagecrl") // CPUUSAGE_CRL;
    $self->{cpuusagecrl} = is_int($cpuusagecrl) ? $cpuusagecrl : CPUUSAGE_CRL;

    # LoadAVG5WRL & LoadAVG5CRL
    my $loadavgwrl = value($self->ctk->config(), "loadavgwrl") // LOADAVG5_WRL;
    $self->{loadavgwrl} = is_flt($loadavgwrl) ? $loadavgwrl : LOADAVG5_WRL;
    my $loadavgcrl = value($self->ctk->config(), "loadavgcrl") // LOADAVG5_CRL;
    $self->{loadavgcrl} = is_flt($loadavgcrl) ? $loadavgcrl : LOADAVG5_CRL;

    return $self;
}
sub init {
    my $self = shift;
    my $logger = $self->logger; # Logger realy is exists?

    # Get availabled commands
    my @availabled_commands = ();
    foreach my $h (get_list_handlers()) {
        if ($self->{command} eq "default") {
            push @availabled_commands, $h->{name};
        } else {
            push @availabled_commands, $h->{name} if $self->{command} eq $h->{name};
        }
    }
    $self->{availabled_commands} = [@availabled_commands];
    unless (@availabled_commands) {
        $logger ? $logger->log_error("No available handlers") : print STDERR "No available handlers\n";
        return $self->interrupt(1);
    }

    # Allocation commands
    my @allocated_commands = _allocate($self->{forkers}, [@availabled_commands]); # 3, [qw/a1 a2 a3 a4 a5 a6 a7/]
    $self->{allocated_commands} = [@allocated_commands];

    #$logger->log_error(VISD::explain([ @allocated_commands ]));
    #$logger->log_error( "Cmd: %s; Forks: %d", $self->{command}, $self->{forkers} );

    #$logger->log_error( "Redis version: %s", $REDIS_VERSION );
    #$logger->log_error( $self->serializer->serialize( {foo=>1, bar=> "test"} ) );

    #print VISD::explain($self);
    #print VISD::explain($self->logger); $self->interrupt(1);

    # Get POSIX uname data
    my %uname = get_uname();
    $self->{kversion} = sprintf("%s %s %s", $uname{sysname} // '', $uname{release} // '', $uname{version} // '');
    $self->{kmachine} = $uname{machine} || 'unknown';

    # Get CPUinfo
    my %cpuinfo = get_cpuinfo();
    $self->{cpuinfo} = {%cpuinfo};

    if (my $redis = $self->redis) {
        # Kernel version & machine
        $redis->set($self->genkeyf("kversion"), $self->{kversion});
        $redis->set($self->genkeyf("kmachine"), $self->{kmachine});

        # CPU info
        $redis->set($self->genkeyf("cpuinfo"), $self->serializer->serialize( $self->{cpuinfo} ));
    } else {
        return $self->interrupt(1); # Redis required for Cetus working
        #$redis->setex("test.bar", 5, scalar(localtime(time())));
    }

    ##
    ## Working dirs initializing
    ##
    my $local_dir = File::Spec->catdir($self->ctk->datadir(), $self->{hostname});
    preparedir($local_dir);
    my $ahosts = $self->{ahosts};
    foreach my $ahost (@$ahosts) { # AHosts
        preparedir(File::Spec->catdir($self->ctk->datadir(), lc($ahost)));
    }

    #printf STDERR "Localdir: %s\n", $local_dir;
    #return $self->interrupt(1);

    #print VISD::explain($self->{ahosts}); return $self->interrupt(1);
    #print VISD::explain(node($self->ctk->config("host"), "foo")); return $self->interrupt(1);

    # Set fetch timeout
    VISD::Tools::set_fetch_timeout($self->ctk->config("timeout"));

    return 1;
}
sub error {
    my $self = shift;
    my $err = shift;
    return $self->{error} unless defined $err;
    $self->{error} = $err;
    return $self->{error};
}
sub serializer { shift->{serializer} }
sub genkeyf {
    my $self = shift;
    my $key = pop // "incorrect_key_name";
    my $section = pop // "general";
    return sprintf("%s.%s.%s",
        VISD_REDIS_KEY_PFX, $section, $key
    );
}
sub genhkeyf {
    my $self = shift;
    my $hst = shift || 'nan';
    my $key = pop // "incorrect_key_name";
    my $section = pop // "general";
    return sprintf("%s.%s.%s.%s",
        VISD::PROJECTNAMEL, lc($hst), $section, $key
    );
}
sub interfaces {
    my $self = shift;
    my @ifaces = @_;
    return () unless scalar(@ifaces);
    my $allowed = $self->{allowedinterfaces};
    return @ifaces unless $allowed && scalar(@$allowed);
    return (duplicates(@$allowed, @ifaces));
}
sub devices {
    my $self = shift;
    my @devs = @_;
    return () unless scalar(@devs);
    my $allowed = $self->{alloweddevices};
    return @devs unless $allowed && scalar(@$allowed);
    return (duplicates(@$allowed, @devs));
}
sub redis {
    my $self = shift;
    my $ctk = $self->get_ctk;
    my $logger = $self->logger();
    my $redis = $self->{redis};

    my $config = $ctk->config;
    my $redis_server = value($config => lc('RedisServer')) || VISD::REDIS_SERVER;

    # Init/reinit redis client
    unless ($redis) {
        if ($REDIS_VERSION) {
            my $redis_cnct_to = fv2zero(value($config => lc('RedisConnectTimeout'))) || VISD::REDIS_CONNECT_TO;
            my $redis_rqst_to = fv2zero(value($config => lc('RedisRequestTimeout'))) || VISD::REDIS_REQUEST_TO;
            $redis = Redis->new(
                server          => $redis_server,
                name            => "visd_cetus_agent",
                conservative_reconnect => 1,
                cnx_timeout     => $redis_cnct_to,
                read_timeout    => $redis_rqst_to,
                write_timeout   => $redis_rqst_to,
                no_auto_connect_on_new => 1,
            );
            eval {
                $redis->connect();
            };
            if ($@) {
                if ($logger) {
                    $logger->log_error("Can't connect to Redis (%s): %s", $redis_server, $@);
                } else {
                    printf STDERR "Can't connect to Redis (%s): %s", $redis_server, $@;
                }
            }
            $self->{redis} = $redis;
        }
    }
    if ($redis) {
        unless ($redis->ping) {
            if ($logger) {
                $self->logger->log_error("Redis is offline (%s)", $redis_server);
            } else {
                printf STDERR "Redis is offline (%s)", $redis_server;
            }
            $redis = undef;
            $self->{redis} = $redis;
        }
    }

    return $redis;
}
sub get_remote_data {
    my $self = shift;
    my $ahost = shift || DEFAULT_HOSTNAME;
    my $expire = shift;
    my $ctk = $self->get_ctk;
    my $logger = $self->logger();
    my $ldf = File::Spec->catfile($ctk->datadir(), lc($ahost), VISD::LAST_DATA_FILE);
    my $url = uv2null(value($ctk->config("host"), $ahost, "sourceurl"));

    # Fetching data
    my ($status, $msg) = fetch($url, File::Spec->catdir($ctk->datadir(), lc($ahost)));
    unless ($status) { # Error
        $logger->log_error($msg);
        return;
    }

    # DeSerialize data
    my $data = $self->serializer->deserialize($msg);
    unless ($self->serializer->status) {
        $logger->log_error($self->serializer->error);
        return;
    }

    # Check host name
    my $hn = uv2null(value($data, "hostname"));
    if ($hn ne $ahost) {
        $logger->log_error("Received host is mismatch! Got: %s; Expected: %s", $hn, $ahost);
        return;
    }

    return $data;
}
sub get_status { # type, data
    my $self = shift;
    my $type = shift || ''; # general, sys, net, dev
    my $data = shift || return STATUS_UNKNOWN;
    my $status = STATUS_UNKNOWN;
    # System health status:
    # -2  Unknown
    # -1  Warning
    #  0  Critical
    #  1  Good

    my $degrade_status = sub {
        my $st = shift;
        if ($st == STATUS_GOOD) {
            $status = $st if $status == STATUS_UNKNOWN;
        } elsif ($st == STATUS_WARNING) {
            $status = $st if $status != STATUS_CRITICAL;
        } else {
            $status = $st;
        }
        return $status;
    };

    if ($type eq 'net') {
        foreach my $v (values(%$data)) {
            next if $status == STATUS_CRITICAL;
            if (value($v, "receive") && value($v, "transmit")) { # Ok
                &$degrade_status(STATUS_GOOD);
            } else { # Error
                if ($status == STATUS_WARNING) {
                    $status = STATUS_CRITICAL;
                } else {
                    $status = STATUS_WARNING;
                }
            }
        }
    } elsif ($type eq 'dev') {
        foreach my $v (values(%$data)) {
            next if $status == STATUS_CRITICAL;
            my $total = fv2zero(value($v, "total"));
            unless ($total) {
                $status = STATUS_WARNING;
                next;
            }
            my $used = fv2zero(value($v, "used"));
            my $usage = ($used * 100) / $total;
            if (($usage <= $self->{devicewrl}) && ($usage <= $self->{devicecrl})) { # Ok
                &$degrade_status(STATUS_GOOD);
            } elsif ($usage <= $self->{devicecrl}) { # Warning
                $status = STATUS_WARNING;
            } else {
                $status = STATUS_CRITICAL;
            }
        }
    } elsif ($type eq 'sys') {
        # cpuusage
        &$degrade_status(($data->{cpu} > $self->{cpuusagewrl}) && ($data->{cpu} <= $self->{cpuusagecrl})
            ? STATUS_WARNING
            : ($data->{cpu} > $self->{cpuusagecrl} ? STATUS_CRITICAL : STATUS_GOOD)
        );

        # loadavg
        my $loadavg = $data->{avg5} / $data->{procs}; # AVG5 per one processor
        &$degrade_status(($loadavg > $self->{loadavgwrl}) && ($loadavg <= $self->{loadavgcrl})
            ? STATUS_WARNING
            : ($loadavg > $self->{loadavgcrl} ? STATUS_CRITICAL : STATUS_GOOD)
        );

        # mem, swp, iod -- is temporary skipped!
    } else {
        foreach my $s (qw/status_sys status_net status_dev/) {
            &$degrade_status($data->{$s} // STATUS_UNKNOWN);
        }
    }

    return $status;
}

# AnyEvent process with watcher support
sub run {
    my $self = shift;
    my $ctk = $self->get_ctk;
    my $logger = $self->logger;
    my $quit_program = AnyEvent->condvar;

    # CTK
    my $config = $ctk->config;

    return 1 unless $self->ok;

    # Get assigned commands
    my $allocated_commands = $self->{allocated_commands};
    my $assigned_commands = $allocated_commands->[$self->{workerident} - 1];
    unless (is_array($assigned_commands) && @$assigned_commands) {
        $logger->log_warn("Skipped! No assigned handlers for worker #%d", $self->{workerident});
        return 0;
    }
    #$logger->log_info(VISD::explain({ worker => $self->{workerident}, cmds => $assigned_commands }));

    # Create alone process timer
    my $iteration = 0; # Iteration number
    my $after = $self->{workerident} || START_DELAY;
    my $interval = int(uv2zero($ctk->option("interval"))) || INTERVAL; # Time interval. 1 op per n sec
    my $timer = AnyEvent->timer (after => $after, interval => $interval, cb => sub {
        $quit_program->send(1) unless $self->ok;
        $iteration++;

        foreach my $cmd (@$assigned_commands) {
            my $ret = 0;

            # Lookup handler and run it!
            #my $t_start = gettimeofday();
            if(my $handler = $self->lookup_handler($cmd)) {
                my $handler_interval = fv2zero($handlers{$cmd}{interval});
                my $abhi = int($handler_interval/$interval);
                next if ($handler_interval && $abhi && (($iteration - 1) % $abhi)); # Check interval
                $ret = &$handler( $self, $iteration );
            } else {
                $logger->log_error("Handler \"%s\" not found", $cmd);
                $self->exception(1);
                last;
            }

            # Exceptions
            if (!$ret && $self->error) {
                #my $t_wait = sprintf("%.*f", 4, gettimeofday() - $t_start) * 1;
                $logger->log_error("ERROR %s (W=%d; PID=%d; I=%d): %s", #  at %f secs
                    $cmd,
                    $self->{workerident},
                    $self->{workerpid},
                    $iteration,
                    #$t_wait,
                    $self->error
                );
            }
        }

    });

    # Run!
    $quit_program->recv;

    return 1;
}

sub register_handler {
    my $fake = shift;
    my %h = @_;
    my $name  = lc($h{name} || "default");
    croak("Can't register handler: no handler subroutine") unless $h{handler} && ref($h{handler}) eq 'CODE';
    if (exists($handlers{$name}) && is_hash($handlers{$name})) {
        carp(sprintf("Duplicate name \"%s\" in handlers", $name));
        return 0;
    }
    $handlers{$name} = {%h};
    return 1;
}
sub lookup_handler {
    my $self = shift;
    my $name  = lc(shift || "default");
    return undef unless exists($handlers{$name}) && is_hash($handlers{$name});
    my $node = $handlers{$name};
    return undef unless exists($node->{handler}) && ref($node->{handler}) eq 'CODE';
    return $node->{handler};
}
sub get_list_handlers {
    my @ret;
    foreach my $v (values %handlers) {
        my %hv = ();
        foreach my $k (keys %$v) {
            next if $k eq 'handler';
            $hv{$k} = $v->{$k};
        }
        push @ret, {%hv};
    }
    return sort {$a->{order} <=> $b->{order}} @ret;
}

sub _allocate { # 3, [qw/a1 a2 a3 a4 a5 a6 a7/]
    my $forks = shift || 0;
    my $_acts = shift || [];
    my @acts = @$_acts;
    my @res = ();
    while ($forks && @acts) {
        foreach my $f (0..($forks-1)) {
            $res[$f] = [] unless exists $res[$f];
            my $nxt = shift(@acts);
            last unless defined $nxt;
            push @{$res[$f]}, $nxt;
        }
    }
    return @res;
}

1;

__END__
